#define MODULE "exec"

#include "errno.h"
#include "interrupts/Interrupts.h"
#include "kassert.h"
#include "memory/MemoryManager.h"
#include "memory/PMM.h"
#include "memory/VMM.h"
#include "misc/hang.h"
#include "std/stdlib.h"
#include "std/string.h"
#include "sys/Syscall.h"
#include "sys/elf/ELFLoader.h"
#include "thread/Scheduler.h"
#include "utils/Addresses.h"

void sys_fork(Context* context)
{
    kinfoln("fork(): attempting fork");

    Task* parent = Scheduler::current_task();

    Task* child = Scheduler::create_user_task();

    memcpy(&child->regs, &parent->regs, sizeof(Context));
    if (parent->floating_saved)
    {
        memcpy(child->floating_region, parent->floating_region, sizeof(parent->floating_region));
        child->floating_saved = true;
    }
    for (int i = 0; i < TASK_MAX_FDS; i++) { child->files[i] = parent->files[i]; }

    child->address_space = parent->address_space.clone();

    child->regs.rsp += sizeof(uintptr_t) * 2;

    child->regs.rax = 0;
    context->rax = child->id;

    child->state = child->Sleeping;
    child->task_sleep = 1000;

    kinfoln("fork(): parent RIP %lx, child RIP %lx, parent RSP %lx, child RSP %lx", parent->regs.rip, child->regs.rip,
            parent->regs.rsp, child->regs.rsp);

    kinfoln("fork(): forked parent %ld into child %ld", parent->id, child->id);

    return;
}

void sys_exec(Context* context, const char* pathname)
{
    char* kpathname = Syscall::strdup_from_user(pathname);
    if (!kpathname)
    {
        context->rax = -EFAULT;
        return;
    }

    kinfoln("exec(): executing %s", kpathname);

    VFS::Node* program = VFS::resolve_path(kpathname);
    if (!program)
    {
        kfree(kpathname);
        context->rax = -ENOENT;
        return;
    }

    if (program->type == VFS_DIRECTORY)
    {
        kfree(kpathname);
        context->rax = -EISDIR;
        return;
    }

    long memusage;
    if ((memusage = ELFLoader::check_elf_image(program)) < 0)
    {
        kfree(kpathname);
        context->rax = -ENOEXEC;
        return;
    }

    if ((uint64_t)memusage > PMM::get_free())
    {
        kfree(kpathname);
        context->rax = -ENOMEM;
        return;
    }

    Interrupts::disable();
    ASSERT(!Interrupts::are_enabled()); // This part is pretty sensitive.

    Task* task = Scheduler::current_task();
    ASSERT(task);

    // At this point, pretty much nothing can fail.

    ELFImage* image = ELFLoader::load_elf_from_vfs(program);
    ASSERT(image); // If check_elf_image succeeded, load_elf_from_vfs MUST succeed, unless something has gone terribly
                   // wrong.

    Scheduler::reset_task(task, image);

    set_context_from_task(*task, context);

    kfree(kpathname);

    return;
}